<?php

namespace Drupal\advanced_sanitize;

use Drupal\advanced_sanitize\Event\AdvancedSanitizeEvents;
use Drupal\advanced_sanitize\Event\FinishedSanitizeEvent;
use Drupal\advanced_sanitize\Event\PostSanitizeEvent;
use Drupal\advanced_sanitize\Event\PostSanitizeRevisionEvent;
use Drupal\advanced_sanitize\Event\PreSanitizeEvent;
use Drupal\advanced_sanitize\Event\PreSanitizeRevisionEvent;
use Drupal\Component\Serialization\Yaml;
use Drupal\Core\Batch\BatchBuilder;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\DependencyInjection\DependencySerializationTrait;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeBundleInfo;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Faker\Factory;
use Faker\Generator;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;

/**
 * Service description.
 */
class AdvancedSanitizeService {

  use StringTranslationTrait;
  use DependencySerializationTrait;

  /**
   * Default bundle id.
   *
   * @var string
   */
  protected const PLACEHOLDER_BUNDLE = 'default';

  /**
   * Processor ids being recognized by module.
   *
   * @var string
   */
  protected const ALLOWED_DATA_PROVIDERS = ['faker', 'constant', 'sql'];

  /**
   * Singular batch operation item limit.
   *
   * @var int
   */
  protected const LIMIT = 1;

  /**
   * Faker factory.
   *
   * @var \Faker\Generator
   */
  protected Generator $faker;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfo
   */
  protected $bundleInfo;

  /**
   * The logger channel factory.
   *
   * @var \Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher
   */
  protected LoggerChannelInterface $logger;

  /**
   * Event dispatcher service.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected EventDispatcherInterface $eventDispatcher;

  /**
   * Advanced sanitize configs.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * Constructs an AdvancedSanitizeService object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Database\Connection $database
   *   The database connection.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   * @param \Drupal\Core\Entity\EntityTypeBundleInfo $bundle_info
   *   Bundle info.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger
   *   The logger channel factory.
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $event_dispatcher
   *   Event dispatcher service.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager,
                              Connection $database,
                              EntityFieldManagerInterface $entity_field_manager,
                              ModuleHandlerInterface $module_handler,
                              EntityTypeBundleInfo $bundle_info,
                              LoggerChannelFactoryInterface $logger,
                              EventDispatcherInterface $event_dispatcher,
                              ConfigFactoryInterface $config_factory) {
    $this->faker = Factory::create();
    $this->entityTypeManager = $entity_type_manager;
    $this->database = $database;
    $this->entityFieldManager = $entity_field_manager;
    $this->moduleHandler = $module_handler;
    $this->bundleInfo = $bundle_info;
    $this->logger = $logger->get('advanced_sanitize');
    $this->config = $config_factory->get('advanced_sanitize.settings');
    $this->eventDispatcher = $event_dispatcher;
  }

  /**
   * Discovers and validates sanitize config.
   *
   * @return array
   *   Definitions array to process.
   */
  public function retrieveConfig(): array {
    $realpath = $this->config->get('config_path');
    if (!empty($realpath)) {
      $sanitizeConfig = Yaml::decode(file_get_contents($realpath));
      foreach ($sanitizeConfig as $key => $sanitizeRecord) {
        // Checking "sql" records.
        if (array_key_exists('table_name', $sanitizeRecord) &&
          !array_key_exists('entity_id', $sanitizeRecord)) {
          $this->checkSqlRecord($sanitizeRecord, $sanitizeConfig, $key);
        }
        // Checking "entity" records.
        if (array_key_exists('entity_id', $sanitizeRecord) &&
          !array_key_exists('table_name', $sanitizeRecord)) {
          $this->checkEntityRecord($sanitizeRecord, $sanitizeConfig, $key);
        }

      }
      return array_filter($sanitizeConfig);
    }
    return [];
  }

  /**
   * Checks sql-related sanitize record before processing.
   */
  protected function checkSqlRecord(array $sqlRecord, array &$sanitizeConfig, int $key): void {
    if (!array_key_exists('column_mapping', $sqlRecord)) {
      $this->logger->warning('Sql record %key contains not enough column data!', [
        '@key' => $key,
      ]);
      unset($sanitizeConfig[$key]);
    }
  }

  /**
   * Checks entity-related sanitize record before processing.
   *
   * @param array $entitySanitizeRecord
   *   Single sanitize record.
   * @param array $sanitizeConfig
   *   Sanitize configuration.
   * @param int $key
   *   Record id within sanitize config.
   */
  protected function checkEntityRecord(array $entitySanitizeRecord, array &$sanitizeConfig, int $key): void {
    $entityId = $entitySanitizeRecord['entity_id'];
    $bundleId = $entitySanitizeRecord['bundle_id'];
    $fieldMapping = $entitySanitizeRecord['field_mapping'];
    if (!empty($fieldMapping)) {
      // Load the field definitions for a bundle.
      if ($bundleId === self::PLACEHOLDER_BUNDLE) {
        $bundleFields = $this->entityFieldManager->getFieldStorageDefinitions($entityId);
      }
      else {
        $bundleFields = $this->entityFieldManager->getFieldDefinitions($entityId, $bundleId);
      }
      $bundleFields = array_keys($bundleFields);

      // Check field mapping.
      foreach ($fieldMapping as $fieldKey => $fieldDefinition) {
        // Verify that field exists within entity.
        $fieldId = $fieldDefinition['field_id'];
        if (!in_array($fieldId, $bundleFields, TRUE)) {
          $this->logger->warning('Field @field_id was not found within @entity_id entity definition!', [
            '@field_id' => $fieldId,
            '@entity_id' => $entityId,
          ]);
          unset($sanitizeConfig[$key]['field_mapping'][$fieldKey]);
        }
        // Validate data providers.
        if (!in_array($fieldDefinition['data_provider'], self::ALLOWED_DATA_PROVIDERS, TRUE)) {
          $this->logger->warning('Unsupported data provider: @processor', [
            '@processor' => $fieldDefinition['data_provider'],
          ]);
          // Remove config entry of not-supported data providers.
          unset($sanitizeConfig[$key]['field_mapping'][$fieldKey]);
        }
      }
    }
    else {
      $this->logger->warning('Sanitize configuration is empty for @entity entity (@bundle bundle)!', [
        '@entity' => $entityId,
        '@bundle' => $bundleId,
      ]);
      unset($sanitizeConfig[$key]);
    }
  }

  /**
   * Sets up batch processing.
   *
   * @param bool $runInBackground
   *   With TRUE sets batch for launch from backend.
   */
  public function sanitize(bool $runInBackground = TRUE): void {
    $validDefinitions = $this->retrieveConfig();
    if (!empty($validDefinitions)) {
      $batchBuilder = $this->setBatchBuilder($validDefinitions);
      batch_set($batchBuilder->toArray());
      if ($runInBackground) {
        drush_backend_batch_process();
      }
    }
  }

  /**
   * Sets a batch builder.
   *
   * @param array $validDefinitions
   *   Sanitize definitions to process.
   */
  public function setBatchBuilder(array $validDefinitions): BatchBuilder {
    $batchBuilder = new BatchBuilder();
    $batchBuilder->setTitle($this->t('Processing...'))
      ->setInitMessage($this->t('Initializing...'))
      ->setProgressMessage($this->t('Sanitized @current of @total.'))
      ->setErrorMessage($this->t('An error has occurred.'))
      ->addOperation([$this, 'processBatch'], [$validDefinitions])
      ->setFinishCallback([$this, 'finishBatch']);

    return $batchBuilder;
  }

  /**
   * Process single batch operation.
   *
   * @param array $definitions
   *   Valid entity definitions.
   * @param array|object $context
   *   Batch context.
   */
  public function processBatch(array $definitions, &$context): void {
    if (empty($context['sandbox']['progress'])) {
      $total = array_keys($definitions);

      $context['sandbox']['total'] = count($total);
      $context['sandbox']['progress'] = 0;
    }

    if (empty($context['sandbox']['items'])) {
      $context['sandbox']['items'] = $definitions;
    }

    $counter = 0;
    if (!empty($context['sandbox']['items'])) {
      if ($context['sandbox']['progress'] != 0) {
        array_splice($context['sandbox']['items'], 0, self::LIMIT);
      }

      foreach ($context['sandbox']['items'] as $entityDefinition) {
        if ($counter !== self::LIMIT) {

          $context['message'] = $this->t('Now processing @importableEntityId :progress of :count', [
            '@importableEntityId' => $entityDefinition['entity_id'] ?? $entityDefinition['table_name'],
            ':progress' => $context['sandbox']['progress'] + 1,
            ':count' => $context['sandbox']['total'],
          ]);

          $this->processBatchItem($entityDefinition);

          $counter++;
          $context['sandbox']['progress']++;

          $context['results']['processed'] = $context['sandbox']['progress'];
        }
      }
    }

    if ($context['sandbox']['progress'] !== $context['sandbox']['total']) {
      $context['finished'] = $context['sandbox']['progress'] / $context['sandbox']['total'];
    }
  }

  /**
   * Handles sanitization of single entity definition passed to batch.
   *
   * @param array $sanitizeRecord
   *   Processed sanitize record.
   */
  public function processBatchItem(array $sanitizeRecord): void {
    switch (TRUE) {
      case array_key_exists('table_name', $sanitizeRecord):
        $this->processWithSql($sanitizeRecord);
        break;

      case array_key_exists('entity_id', $sanitizeRecord):
        $this->processWithDrupal($sanitizeRecord);
        break;
    }
  }

  /**
   * Handles field data sanitization with Faker.
   *
   * @param array $sanitizeRecord
   *   Processed sanitize record.
   */
  public function processWithDrupal(array $sanitizeRecord): void {
    foreach ($sanitizeRecord['field_mapping'] as $fieldMapping) {
      $query = $this->entityTypeManager->getStorage($sanitizeRecord['entity_id'])->getQuery();
      if ($sanitizeRecord['bundle_id'] !== self::PLACEHOLDER_BUNDLE) {
        $query->condition('type', $sanitizeRecord['bundle_id']);
      }
      // Disabling access check to run correctly from drush.
      $query->accessCheck(FALSE);
      $entityIds = $query->execute();

      $entityType = $this->entityTypeManager->getStorage($sanitizeRecord['entity_id'])->getEntityType();
      $entityTypeRevisionable = FALSE;
      if ($entityType->isRevisionable()) {
        $entityTypeRevisionable = TRUE;
      }

      foreach ($entityIds as $entityId) {
        $entity = $this->entityTypeManager->getStorage($sanitizeRecord['entity_id'])->load($entityId);
        if (isset($entity)) {
          $event = new PreSanitizeEvent($entity);
          $this->eventDispatcher->dispatch($event, AdvancedSanitizeEvents::PRE_SANITIZE);
          // Leave empty fields empty.
          if (!empty($entity->get($fieldMapping['field_id'])->value)) {
            switch ($fieldMapping['data_provider']) {
              case 'faker':
                $entity->set($fieldMapping['field_id'],
                  $this->getFakerData(
                    $fieldMapping['method'],
                    $fieldMapping['parameters'] ?? NULL
                  ));
                break;

              case 'constant':
                $entity->set(
                  $fieldMapping['field_id'],
                  $fieldMapping['value']
                );
                break;
            }
            $entity->save();

            $event = new PostSanitizeEvent($entity);
            $this->eventDispatcher->dispatch($event, AdvancedSanitizeEvents::POST_SANITIZE);

            if ($entityTypeRevisionable) {
              $this->sanitizeRevisions($entity);
            }
          }

        }
      }
    }

  }

  /**
   * Handles field data replacement with sql expression.
   *
   * @param array $sanitizeRecord
   *   Processed entity's field mapping.
   */
  public function processWithSql(array $sanitizeRecord): void {
    $query = $this->database->update($sanitizeRecord['table_name']);
    $updFields = [];
    if (array_key_exists('where_expr', $sanitizeRecord)) {
      $query->where($sanitizeRecord['where_expr'], $sanitizeRecord['where_args'] ?? []);
    }
    foreach ($sanitizeRecord['column_mapping'] as $colSanitizeDefinition) {
      switch (TRUE) {
        case $colSanitizeDefinition['data_provider'] === 'faker':
          $updFields[$colSanitizeDefinition['col_name']] =
            $this->getFakerData(
            $colSanitizeDefinition['method'],
            $colSanitizeDefinition['parameters']
          );
          break;

        case $colSanitizeDefinition['data_provider'] === 'constant':
          $updFields[$colSanitizeDefinition['col_name']] =
            $colSanitizeDefinition['value'];
          break;

        case $colSanitizeDefinition['data_provider'] === 'expression':
          $query
            ->expression(
              $colSanitizeDefinition['col_name'],
              $colSanitizeDefinition['expression'],
              $colSanitizeDefinition['expr_arguments']
          );
          break;
      }

    }
    if (!empty($updFields)) {
      $query->fields($updFields);
    }
    $query->execute();
  }

  /**
   * Make all revisions copy given already sanitized entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Sanitized entity.
   */
  public function sanitizeRevisions(EntityInterface $entity): void {
    $revisionIds = $this->entityTypeManager
      ->getStorage($entity->getEntityTypeId())
      ->getQuery()
      ->accessCheck()
      ->allRevisions()
      ->condition('type', $entity->getEntityType()->getKey('id'))
      ->execute();

    foreach ($revisionIds as $revisionId) {
      $revision = $this->entityTypeManager->getStorage($entity->getEntityTypeId())->loadRevision($revisionId);

      if (isset($revision)) {
        $event = new PreSanitizeRevisionEvent($revision);
        $this->eventDispatcher->dispatch($event, AdvancedSanitizeEvents::PRE_SANITIZE_REVISION);

        foreach ($entity->getFields() as $name => $field) {
          $revision->set($name, $field->getValue());
        }
        $revision->save();

        $event = new PostSanitizeRevisionEvent($revision);
        $this->eventDispatcher->dispatch($event, AdvancedSanitizeEvents::POST_SANITIZE_REVISION);
      }
    }
  }

  /**
   * Batch finished callback for sanitize operation.
   *
   * @param bool $success
   *   TRUE if batch successfully completed.
   * @param array $results
   *   Batch results.
   * @param array $operations
   *   An array of function calls.
   */
  public function finishBatch(bool $success, array $results, array $operations) {
    drupal_flush_all_caches();
    $finishedSanitiseEvent = new FinishedSanitizeEvent($success, $results, $operations);
    $this->eventDispatcher->dispatch($finishedSanitiseEvent, AdvancedSanitizeEvents::FINISHED_SANITIZE);
  }

  /**
   * Retrieve fake data from Faker.
   *
   * @param string $method
   *   Faker method to call.
   * @param array|null $parameters
   *   Parameters to use with method.
   *
   * @return mixed
   *   Data generated by Faker.
   */
  protected function getFakerData(string $method, array|null $parameters): mixed {
    return $this->faker
      ->unique()
      ->format(
      $method,
      $parameters ?? []
    );
  }

}
